/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** project:
**    OpenGL/cairo anti-aliasing trick
**
** file:
**    cairo-utils.c
**
** author:
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**
** copyright (C) Mirco Mueller, December 2006, placed under the terms of the GPL
**
*******************************************************************************/

#include <math.h>
#include <librsvg/rsvg.h>
#include <librsvg/rsvg-cairo.h>

#include "cairo-utils.h"

/* just draw a path resembling a rectangle with rounded corners; there are no
** checks on width, height and corner-radius performed */
void
cairo_draw_round_rect (cairo_t* pContext,
		       double fX,            /* top-left corner         */
		       double fY,            /* top-left corner         */
		       double fCornerRadius, /* "size" of the corners   */
		       double fWidth,        /* width of the rectangle  */
		       double fHeight        /* height of the rectangle */)
{
	/* top-left, right of the corner */
	cairo_move_to (pContext, fX + fCornerRadius, fY);

	/* top-right, left of the corner */
	cairo_line_to (pContext,
		       fX + fWidth - fCornerRadius,
		       fY);

	/* top-right, below the corner */
	cairo_curve_to (pContext,
			fX + fWidth, fY,
			fX + fWidth, fY,
			fX + fWidth, fY + fCornerRadius);

	/* bottom-right, above the corner */
	cairo_line_to (pContext,
		       fX + fWidth,
		       fY + fHeight - fCornerRadius);

	/* bottom-right, left of the corner */
	cairo_curve_to (pContext,
			fX + fWidth, fY + fHeight,
			fX + fWidth, fY + fHeight,
			fX + fWidth - fCornerRadius, fY + fHeight);

	/* bottom-left, right of the corner */
	cairo_line_to (pContext,
		       fX + fCornerRadius,
		       fY + fHeight);

	/* bottom-left, above the corner */
	cairo_curve_to (pContext,
			fX, fY + fHeight,
			fX, fY + fHeight,
			fX, fY + fHeight - fCornerRadius);

	/* top-left, below the corner */
	cairo_line_to (pContext,
		       fX,
		       fY + fCornerRadius);

	/* top-left, right of the corner */
	cairo_curve_to (pContext,
			fX, fY,
			fX, fY,
			fX + fCornerRadius, fY);
}

gint
to_int (gfloat fValue)
{
	fValue = floor (fValue * 256);

	if (fValue > 255)
		return 255;
	else
		return (gint) fValue;
}

gfloat
d1 (gint iValue)
{
	return iValue / 255.0f;
}

gfloat
d2 (gint iValue)
{
	return (iValue + 0.5f) / 256.0f;
}

void
demultiply_alpha (guchar* pucPixelBuffer, /* pointer to image-data */
		  gint iWidth,            /* width of image        */
		  gint iHeight,           /* height of image       */
		  gint iStride            /* stride of image       */)
{
	gint iX = 0;
	gint iY = 0;

	/* sanity check */
	g_assert (pucPixelBuffer != NULL);

	/* get rid of premultiplied-alpha for use in OpenGL */
	for (iX = 0; iX < iWidth; iX++)
	{
		for (iY = 0; iY < iHeight; iY++)
		{
			guchar ucRed;
			guchar ucGreen;
			guchar ucBlue;
			guchar ucAlpha;

			ucRed = pucPixelBuffer[iY * iStride + iX * 4 + 2];
			ucGreen = pucPixelBuffer[iY * iStride + iX * 4 + 1];
			ucBlue = pucPixelBuffer[iY * iStride + iX * 4 + 0];
			ucAlpha = pucPixelBuffer[iY * iStride + iX * 4 + 3];

			ucRed	= to_int (d2 (ucRed) / d1 (ucAlpha));
			ucGreen	= to_int (d2 (ucGreen) / d1 (ucAlpha));
			ucBlue	= to_int (d2 (ucBlue) / d1 (ucAlpha));

			pucPixelBuffer[iY * iStride + iX * 4 + 0] = ucBlue;
			pucPixelBuffer[iY * iStride + iX * 4 + 1] = ucGreen;
			pucPixelBuffer[iY * iStride + iX * 4 + 2] = ucRed;
		}
	}
}

/* one of my special tricks for smooth solid looking graphics, antialising
** without the actual need for aa-hardware, thus very fast and still looking
** perfect; the idea is to leverage the texture-filtering hardware of GL-cards
** to smooth out the edges of a specially prepared texture; slightly trim the
** edges of a texture using cairo's image-surfaces and while doing that also add
** some nicely rounded corners; thanks to cairo's antialised vector-drawing the
** edges are faded out in a subtle way, no hard jaggies; this should be applied
** _before_ the image-data is actually transfered to GL as texture; the supplied
** image-data has to be R8G8B8A8; it only works (or makes sense) for single
** quads/rects/etc */
void
cairo_trim_image (guchar* pucPixelBuffer, /* pointer to image-data */
		  gint iWidth,            /* width of image        */
		  gint iHeight,           /* height of image       */
		  gint iStride            /* stride of image       */)
{	
	cairo_surface_t* pSurface = NULL;
	cairo_status_t status = CAIRO_STATUS_SUCCESS;
	cairo_t* pContext = NULL;
	cairo_pattern_t* pPattern = NULL;

	/* sanity check */
	g_assert (pucPixelBuffer != NULL);

	/* create surface from supplied texture-data */
	pSurface = cairo_image_surface_create_for_data (pucPixelBuffer,
							CAIRO_FORMAT_ARGB32,
							iWidth,
							iHeight,
							iStride);

	/* check for errors */
	status = cairo_surface_status (pSurface);
	if (status != CAIRO_STATUS_SUCCESS)
	{
		g_print ("%s\n", cairo_status_to_string (status));
		return;
	}

	/* create the rendering context */
	pContext = cairo_create (pSurface);

	/* check for any errors */
	status = cairo_status (pContext);
	if (status != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (pSurface);
		g_print ("%s\n", cairo_status_to_string (status));
		return;
	}

	/* ensure that we can use normalized coordinates */
	cairo_scale (pContext, (double) iWidth, (double) iHeight);

	/* create a pattern */
	pPattern = cairo_pattern_create_rgba  (0.0f, 0.0f, 0.0f, 1.0f);
	status = cairo_pattern_status (pPattern);
	if (status != CAIRO_STATUS_SUCCESS)
	{
		cairo_destroy (pContext);
		cairo_surface_destroy (pSurface);
		g_print ("%s\n", cairo_status_to_string (status));
		return;
	}

	/* now use the pattern to draw a pattern-filled rounded rectangle */
	cairo_set_operator (pContext, CAIRO_OPERATOR_DEST_IN);
	cairo_set_source (pContext, pPattern);
	cairo_draw_round_rect (pContext, 0.01f, 0.01f, 0.15f, 0.98f, 0.98f);	
	cairo_fill (pContext);

	/* get rid of premultiplied-alpha for use in OpenGL */
	demultiply_alpha (pucPixelBuffer, iWidth, iHeight, iStride);

	/* clean up */
	cairo_pattern_destroy (pPattern);
	cairo_destroy (pContext);
	cairo_surface_destroy (pSurface);
}

Texture*
cairo_text_to_texture (gchar* pcText, /* the text to render              */
		       gint iWidth,   /* width of window text appears in */
		       gint iHeight   /* height of window text appears in*/)
{
	guchar*			pucBufferData = NULL;
	gint			iBufferSize = 0;
	cairo_surface_t*	pTempSurface = NULL;
	cairo_t*		pTempContext = NULL;
	PangoFontDescription*	pDesc = NULL;
	PangoLayout*		pLayout = NULL;
	cairo_surface_t*	pTextSurface = NULL;
	cairo_t*		pTextContext = NULL;
	PangoRectangle		ink = {0, 0, 0, 0};
	PangoRectangle		log = {0, 0, 0, 0};
	Texture*		pTexture = NULL;

	/* sanity check */
	g_assert (pcText != NULL);

	/* setup some buffer for creating a temporary context */
	iBufferSize = iWidth * iHeight * 4;
	pucBufferData = (guchar*) g_slice_alloc0 (iBufferSize);
	if (!pucBufferData)
	{
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not allocate the temporary buffer!\n");
		return NULL;
	}

	/* create surface from supplied buffer and check for errors */
	pTempSurface = cairo_image_surface_create_for_data (pucBufferData,
							    CAIRO_FORMAT_ARGB32,
							    iWidth,
							    iHeight,
							    iWidth * 4);
	if (cairo_surface_status (pTempSurface) != CAIRO_STATUS_SUCCESS)
	{
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create temporary surface!\n");
		return NULL;
	}

	/* create a temporary rendering context and check for errors */
	pTempContext = cairo_create (pTempSurface);
	if (cairo_status (pTempContext) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (pTempSurface);
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not temporary context!\n");
		return NULL;
	}

	/* setup a new pango-layout based on the temporary context */
	pLayout = pango_cairo_create_layout (pTempContext);
	if (!pLayout)
	{
		cairo_destroy (pTempContext);
		cairo_surface_destroy (pTempSurface);
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create pango-layout!\n");
		return NULL;
	}

	/* get a new pango-description */
	pDesc = pango_font_description_new ();
	if (!pDesc)
	{
		g_object_unref (pLayout);
		cairo_destroy (pTempContext);
		cairo_surface_destroy (pTempSurface);
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create pango-font-description!\n");
		return NULL;
	}

	/* set some desired font-attributes */
	pango_font_description_set_absolute_size (pDesc, 24 * PANGO_SCALE);
	pango_font_description_set_family_static (pDesc,
						  "DejaVu LGC Sans");
	pango_font_description_set_weight (pDesc, PANGO_WEIGHT_SEMIBOLD);
	pango_layout_set_font_description (pLayout, pDesc);
	pango_font_description_free (pDesc);

	/* set the text to render and... */
	pango_layout_set_text (pLayout, pcText, -1);

	/* ... get size of rendered string in pixels */
	pango_layout_get_pixel_extents (pLayout, &ink, &log);

	/* get rid of the temporary context and buffer, no longer needed */
	cairo_destroy (pTempContext);
	cairo_surface_destroy (pTempSurface);
	g_slice_free1 (iBufferSize, (gpointer) pucBufferData);

	/* get a buffer for creating the text-context */
	iBufferSize = log.width * log.height * 4;
	pucBufferData = (guchar*) g_slice_alloc0 (iBufferSize);
	if (!pucBufferData)
	{
		g_object_unref (pLayout);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not allocate the temporary buffer!\n");
		return NULL;
	}

	/* create a new surface for the text */
	pTextSurface = cairo_image_surface_create_for_data (pucBufferData,
							    CAIRO_FORMAT_ARGB32,
							    log.width,
							    log.height,
							    log.width * 4);
	if (cairo_surface_status (pTextSurface) != CAIRO_STATUS_SUCCESS)
	{
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_object_unref (pLayout);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create text-surface!\n");
		return NULL;
	}

	/* create a context from that new text surface */
	pTextContext = cairo_create (pTextSurface);
	if (cairo_status (pTextContext) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (pTextSurface);
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_object_unref (pLayout);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create drawing-context for text!\n");
		return NULL;
	}

	/* create our own texture object */
	pTexture = (Texture*) g_slice_alloc0 (sizeof (Texture));
	if (!pTexture)
	{
		cairo_destroy (pTextContext);
		cairo_surface_destroy (pTextSurface);
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_object_unref (pLayout);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not allocate memory for the texture struct!\n");
		return NULL;
	}

	/* clear the newly created text-context */
	cairo_set_operator (pTextContext, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (pTextContext, 0.0f, 0.0f, 0.0f, 0.0f);
	cairo_paint (pTextContext);

	/* draw the text on the the text-context */
	cairo_set_source_rgba (pTextContext, 0.1f, 0.1f, 0.7f, 1.0f);
	cairo_move_to (pTextContext, 0.0f, 0.0f);
	pango_cairo_layout_path (pTextContext, pLayout);
	cairo_fill (pTextContext);

	/* fill the resulting texture-struct */
	pTexture->iTarget = opengl_what_texture_target (log.width, log.height);
	pTexture->fWidth = (float) log.width;
	pTexture->fHeight = (float) log.height;
	pTexture->iBufferSize = iBufferSize;
	pTexture->pucBufferData = pucBufferData; /* to be freed by the caller */

	/* get rid of premultiplied-alpha for use in OpenGL */
	demultiply_alpha (pucBufferData, log.width, log.height, log.width * 4);

	/* free all no longer needed resources */
	cairo_destroy (pTextContext);
	cairo_surface_destroy (pTextSurface);
	g_object_unref (pLayout);

	return pTexture; /* to be freed by the caller */
}

/* TODO: not done yet */
Texture*
cairo_logos_to_texture (gchar* pcSvgFileOne, /* full path first SVG-file  */
			gchar* pcSvgFileTwo, /* full path second SVG-file */
			gint iWidth,	     /* width of window           */
			gint iHeight	     /* height of window          */)
{
	Texture* pTexture = NULL;
	/*RsvgHandle* pSvgOne = NULL;
	RsvgHandle* pSvgTwo = NULL;
	RsvgDimensionData dimensionsOne;
	RsvgDimensionData dimensionsTwo;
	double fAspectOne = 1.0f;
	double fAspectTwo = 1.0f;
	cairo_surface_t* pSurface = NULL;
	cairo_t* pContext = NULL;
	GError* pError = NULL;
	gint iBufferSize = 0;
	guchar* pucBufferData = NULL;

	* sanity checks *
	g_assert (pcSvgFileOne != NULL);
	g_assert (pcSvgFileTwo != NULL);

	* setup some buffer for creating a temporary context *
	iBufferSize = iWidth * iHeight * 4;
	pucBufferData = (guchar*) g_slice_alloc0 (iBufferSize);
	if (!pucBufferData)
	{
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could not allocate the buffer!\n");
		return NULL;
	}

	* create surface from supplied buffer and check for errors *
	pSurface = cairo_image_surface_create_for_data (pucBufferData,
							CAIRO_FORMAT_ARGB32,
							iWidth,
							iHeight,
							iWidth * 4);
	if (cairo_surface_status (pSurface) != CAIRO_STATUS_SUCCESS)
	{
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could not create surface!\n");
		return NULL;
	}

	* create a temporary rendering context and check for errors *
	pContext = cairo_create (pSurface);
	if (cairo_status (pContext) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (pSurface);
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could not context!\n");
		return NULL;
	}

	rsvg_init ();

	* load first SVG-file *
	pSvgOne = rsvg_handle_new_from_file (pcSvgFileOne, &pError);
	if (!pSvgOne = NULL)
	{
		cairo_destroy (pContext);
		cairo_surface_destroy (pSurface);
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could load first SVG-file!\n");
		return NULL;
	}

	* load second SVG-file *
	pSvgTwo = rsvg_handle_new_from_file (pcSvgFileTwo, &pError);
	if (!pSvgTwo)
	{
		rsvg_handle_free (pSvgOne);
		cairo_destroy (pContext);
		cairo_surface_destroy (pSurface);
		g_slice_free1 (iBufferSize, (gpointer) pucBufferData);
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could load second SVG-file!\n");
		return NULL;
	}

	* get dimensions of the two SVG-files *
	rsvg_handle_get_dimensions (pSvgOne, &dimensionsOne);
	rsvg_handle_get_dimensions (pSvgTwo, &dimensionsTwo);

	* sanity check - to avoid division by zero later on *
	if (dimensionsOne.width == 0)
		dimensionsOne.width = 1;

	if (dimensionsOne.height == 0)
		dimensionsOne.height = 1;

	if (dimensionsTwo.width == 0)
		dimensionsTwo.width = 1;

	if (dimensionsTwo.height == 0)
		dimensionsTwo.height = 1;

	* get aspect-ratios of both SVG-files *
	fAspectOne = (double) dimensionsOne.width /
		     (double) dimensionsOne.height;

	fAspectTwo = (double) dimensionsTwo.width /
		     (double) dimensionsTwo.height;

	cairo_scale (pContext,
		     (double) iWidth / (double) dimensionsOne.width,
		     (double) iHeight / (double) dimensionsOne.height);

	rsvg_handle_render_cairo (pSvgOne, pContext);
	rsvg_handle_render_cairo (pSvgTwo, pContext);

	rsvg_handle_free (pSvgOne);
	rsvg_handle_free (pSvgTwo);


	rsvg_term ();*/

	return pTexture;
}
