/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** project:
**    OpenGL/cairo anti-aliasing trick
**
** file:
**    opengl.c
**
** author:
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**
** copyright (C) Mirco Mueller, December 2006, placed under the terms of the GPL
**
*******************************************************************************/

#include <gtk/gtk.h>
#include <gtk/gtkgl.h>
#include <gdk/gdkx.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "shaders.h"
#include "cairo-utils.h"

#define Z_NEAR  1.0f
#define Z_FAR  50.0f
#define FOVY   20.0f

gboolean g_bHasShaders = FALSE;
GLuint g_auiFragProgs[3] = {0, 0, 0};
GLuint g_auiTextureIds[3] = {0, 0, 0};
Texture* g_apTextTextures[3];

typedef GLvoid
(*glXSwapIntervalSGIFunc) (GLint);

typedef GLvoid
(*glXSwapIntervalMESAFunc) (GLint);

/* enable (opengl_enable_vsync (2)) or disable (opengl_enable_vsync (0))
** the synchronization of the GL-rendering to the vblank of the display, for
** nice smooth graphics... you don't want to leave home without it;
** function-pointers are not saved as this call, because it is usually only
** needed once, thus it would not make much sense to keep the pointers around */
void
opengl_enable_vsync (gint iInterval)
{
	const gchar* pcDummy = NULL;
	glXSwapIntervalSGIFunc glXSwapIntervalSGI = 0;
	glXSwapIntervalMESAFunc glXSwapIntervalMESA = 0;

	pcDummy = glXQueryExtensionsString (GDK_DISPLAY_XDISPLAY
		(gdk_display_get_default ()), 0);

	if (g_strrstr (pcDummy, "GLX_SGI_swap_control") != NULL)
	{
		glXSwapIntervalSGI = (glXSwapIntervalSGIFunc)
			gdk_gl_get_proc_address ("glXSwapIntervalSGI");

		if (glXSwapIntervalSGI)
			glXSwapIntervalSGI (iInterval);
		else
			g_print ("Could not get glXSwapIntervalSGI()\n");
	}
	else if (g_strrstr (pcDummy, "GLX_MESA_swap_control") != NULL)
	{
		glXSwapIntervalMESA = (glXSwapIntervalMESAFunc)
			gdk_gl_get_proc_address ("glXSwapIntervalMESA");

		if (glXSwapIntervalMESA)
			glXSwapIntervalMESA (iInterval);
		else
			g_print ("Could not get glXSwapIntervalMESA()\n");
	}
	else
		g_print ("Unable to switch sync-to-vblank on.\n");
}

GLenum
opengl_what_texture_target (GLuint iWidth,
			    GLuint iHeight)
{
	GLenum iTarget = 0;
	gboolean bWidthIsPOT = FALSE;
	gboolean bHeightIsPOT = FALSE;
	gint i = 1;

	/* test width for power-of-two */
	for (i = 1; i <= 32; i++)
		if ((unsigned long) iWidth - (1L << i) == 0)
		{
			bWidthIsPOT = TRUE;
			break;
		}

	/* test height for power-of-two */
	for (i = 1; i <= 32; i++)
		if ((unsigned long) iHeight - (1L << i) == 0)
		{
			bHeightIsPOT = TRUE;
			break;
		}

	/* final conclusion */
	if (bWidthIsPOT && bHeightIsPOT)
		iTarget = GL_TEXTURE_2D;
	else
		iTarget = GL_TEXTURE_RECTANGLE_ARB;

	return iTarget;	
}

/* when you write GL-programs you will make mistakes, there's no doubt about 
** that; to make it a bit easier to identify a possible trouble-maker (e.g.
** something just not renders correctly or at all, but the program does not 
** crash, so gdb will not be of any help) you'll need something different, that
** is what opengl_print_error() is meant for; use it like this:
**
**   ...
**   glGetError ();
**   glTroubleMakerFunc ();
**   opengl_print_error ("glTroubleMakerFunc");
**   ...
**
** the first call clears GL's error-flag, then make your suspicious call and
** finally execute opengl_print_error() and see during runtime what gets spit
** out to the console; for the exact meaning of a reported error you have to
** consult the official GL-documentation (man-pages, red-book, orange-book) */
void
opengl_print_error (gchar* pcFuncName)
{
	GLenum error;

	error = glGetError ();

	switch (error)
	{
		case GL_NO_ERROR :
			g_print ("%s() caused GL_NO_ERROR\n", pcFuncName);
		break;

		case GL_INVALID_ENUM :
			g_print ("%s() caused GL_INVALID_ENUM\n", pcFuncName);
		break;

		case GL_INVALID_VALUE :
			g_print ("%s() GL_INVALID_VALUE\n", pcFuncName);
		break;

		case GL_INVALID_OPERATION :
			g_print ("%s() GL_INVALID_OPERATION\n", pcFuncName);
		break;

		case GL_STACK_OVERFLOW :
			g_print ("%s() GL_STACK_OVERFLOW\n", pcFuncName);
		break;

		case GL_STACK_UNDERFLOW :
			g_print ("%s() GL_STACK_UNDERFLOW\n", pcFuncName);
		break;

		case GL_OUT_OF_MEMORY :
			g_print ("%s() GL_OUT_OF_MEMORY\n", pcFuncName);
		break;

		case GL_TABLE_TOO_LARGE :
			g_print ("%s() GL_TABLE_TOO_LARGE\n", pcFuncName);
		break;
	}
}

/* easily switch between parallel and perspective projection */
void
opengl_change_projection (gboolean bParallel,/* paral. or persp. projection  */
			  GLsizei iWidth,    /* current width of GL-context  */
			  GLsizei iHeight    /* current height of GL-context */)
{
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	if (bParallel)
		glOrtho ((GLdouble) -iWidth / 2, (GLdouble) iWidth / 2,
			 (GLdouble) -iHeight / 2, (GLdouble) iHeight / 2,
			 Z_NEAR, Z_FAR);
	else
		gluPerspective (2.0f * FOVY,
				(GLdouble) iWidth / (GLdouble) iHeight,
				Z_NEAR,
				Z_FAR);

	glMatrixMode (GL_MODELVIEW);
}

/* GL_TEXTURE_3D, GL_PROXY_TEXTURE_CUBE_MAP and mip-mapping are not checked for
** in opengl_texture_proxy_check(), as they are of no interest in this
** application, they can of course be added later on, if the need arises */
gboolean
opengl_texture_proxy_check (GLenum iTarget, /* GL texture-target to test  */
			    gint iChannels, /* number of color-components */
			    gint iFormat,   /* GL image-format to test    */
			    gint iWidth,    /* width of texture to test   */
			    gint iHeight    /* height of texture to test  */)
{
	GLenum iProxyTarget = 0;
	GLint iProxyWidth = 0;
	GLint iProxyHeight = 0;

	/* select appropriate proxy-target */
	switch (iTarget)
	{
		case GL_TEXTURE_1D :
			iProxyTarget = GL_PROXY_TEXTURE_1D;
			glTexImage1D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;

		case GL_TEXTURE_2D :
			iProxyTarget = GL_PROXY_TEXTURE_2D;
			glTexImage2D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      iHeight,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;

		case GL_TEXTURE_RECTANGLE_ARB :
			iProxyTarget = GL_PROXY_TEXTURE_RECTANGLE_ARB;
			glTexImage2D (iProxyTarget,
				      0,
				      iChannels,
				      iWidth,
				      iHeight,
				      0,
				      iFormat,
				      GL_UNSIGNED_BYTE,
				      NULL);
		break;
	}

	/* query the results from the proxy-texture-target */
	glGetTexLevelParameteriv (iProxyTarget,
				  0,
				  GL_TEXTURE_WIDTH,
				  &iProxyWidth);
	glGetTexLevelParameteriv (iProxyTarget,
				  0,
				  GL_TEXTURE_HEIGHT,
				  &iProxyHeight);

	/* do the final check if texture would have fitted within the limits */
	if (iWidth == iProxyWidth &&
	    iHeight == iProxyHeight)
		return TRUE;
	else
		return FALSE;
}

/* very handy to "just quickly" load an image-file from disk and create a GL
** texture-object from it; note though that you have to know if the texture is
** pot or npot and set the iTarget (GL_TEXTURE_1D, GL_TEXTURE_2D... etc.)
** accordingly; for this particular case some special beautifying actions are
** performed on the image in memory... alpha-channel is added if needed and its
** border is trimmed a bit for some nice aa-tricks (that's what's bTrim for) */
GLuint
opengl_upload_texture (GLenum iTarget,    /* GL texture-target to use    */
		       gchar* pcFilename, /* full path/filename of image */
		       gboolean bTrim     /* trim or don't trim edges    */)
{
	GLuint uiTextureId = 0;
	GdkPixbuf* pSrcPixbuf = NULL;
	GdkPixbuf* pNewPixbuf = NULL;
	GError* pError = NULL;
	gint iBitsPerChannel = 0;
	gint iWidth = 0;
	gint iHeight = 0;
	guchar* pucPixelBuffer = NULL;
	gint iChannels = 0;

	/* sanity check */
	g_assert (pcFilename != NULL);

	/* create a GL texture-handle */
	glGenTextures (1, &uiTextureId);
	if (uiTextureId == 0)
		return 0;

	/* load the image from disk */
	pSrcPixbuf = gdk_pixbuf_new_from_file (pcFilename, &pError);
	if (pSrcPixbuf == NULL)
	{
		glDeleteTextures (1, &uiTextureId);
		return 0;
	}

	/* only allow 8 bit per color-component */
	iBitsPerChannel = gdk_pixbuf_get_bits_per_sample (pSrcPixbuf);
	if (iBitsPerChannel != 8)
	{
		g_object_unref (pSrcPixbuf);
		glDeleteTextures (1, &uiTextureId);
		return 0;
	}

	/* add an alpha-channel if not already present and user wants to trim */
	if (!gdk_pixbuf_get_has_alpha (pSrcPixbuf) && bTrim)
		pNewPixbuf = gdk_pixbuf_add_alpha (pSrcPixbuf, FALSE, 0, 0, 0);
	else
		pNewPixbuf = gdk_pixbuf_copy (pSrcPixbuf);

	/* the source pixbuf is no longer needed */
	g_object_unref (pSrcPixbuf);

	/* check if the new pixbuf was created successfully */
	if (pNewPixbuf == NULL)
	{
		glDeleteTextures (1, &uiTextureId);
		return 0;
	}

	/* determine the number of components */
	iChannels = gdk_pixbuf_get_n_channels (pNewPixbuf);

	/* get hold of dimensions and pixel-data */
	iWidth = gdk_pixbuf_get_width (pNewPixbuf);
	iHeight = gdk_pixbuf_get_height (pNewPixbuf);
	pucPixelBuffer = gdk_pixbuf_get_pixels (pNewPixbuf);

	/* see if the texture would fit */
	if (!opengl_texture_proxy_check (iTarget,
					 iChannels,
					 (iChannels == 4) ? GL_RGBA : GL_RGB,
					 iWidth,
					 iHeight))
	{
		g_object_unref (pNewPixbuf);
		glDeleteTextures (1, &uiTextureId);
		return 0;
	}

	/* the MacSlow-ish aa-trick */
	if (bTrim)
		cairo_trim_image (pucPixelBuffer,
				  iWidth,
				  iHeight,
				  iWidth * 4);

	/* finally create the GL texture-object */
	glBindTexture (iTarget, uiTextureId);	
	glTexParameteri (iTarget,
			 GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
	glTexParameteri (iTarget,
			 GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);
	glTexImage2D (iTarget,
		      0,
		      iChannels,
		      iWidth,
		      iHeight,
		      0,
		      (iChannels == 4) ? GL_RGBA : GL_RGB,
		      GL_UNSIGNED_BYTE,
		      pucPixelBuffer);
	g_object_unref (pNewPixbuf);

	return uiTextureId;
}

/*
**
** */
Texture*
opengl_upload_text_texture (gchar* pcText, /* text to render into texture    */
			    GLint iWidth,  /* width of win. text appears in  */
			    GLint iHeight  /* height of win. text appears in */)
{
	GLuint uiTextureId = 0;
	Texture* pTexture = NULL;

	/* sanity check */
	g_assert (pcText != NULL);

	/* get new texture handle and verify its allocation */
	glGenTextures (1, &uiTextureId);
	if (uiTextureId == 0)
		return NULL;

	/* render text into a pixel-buffer via cairo */
	pTexture = cairo_text_to_texture (pcText, iWidth, iHeight);
	if (!pTexture)
	{
		glDeleteTextures (1, &uiTextureId);
		return NULL;
	}

	pTexture->uiTextureId = uiTextureId;

	/* pass pixel-buffer to GL */
	glBindTexture (pTexture->iTarget, pTexture->uiTextureId);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MIN_FILTER,
			 GL_LINEAR);
	glTexParameteri (pTexture->iTarget,
			 GL_TEXTURE_MAG_FILTER,
			 GL_LINEAR);
	glTexImage2D (pTexture->iTarget,
		      0,
		      GL_RGBA,
		      (GLsizei) pTexture->fWidth,
		      (GLsizei) pTexture->fHeight,
		      0,
		      GL_BGRA,
		      GL_UNSIGNED_BYTE,
		      pTexture->pucBufferData);

	/* get rid of resources no longer needed, buffer was allocated in
	** cairo_text_to_texture() */
	g_slice_free1 (pTexture->iBufferSize,
		       (gpointer) pTexture->pucBufferData);

	/* pass valid texture-handle to caller */
	return pTexture;
}

/* just to make the background less boring I've implemented a simple
** checker-board painting function; remember: values to pass differ greatly
** depending on the kind of projection-matrix you've set up */
void
opengl_paint_checker (GLfloat fStartX, /* x-coord of lower-left corner */
		      GLfloat fStartY, /* y-coord of lower-left corner */
		      GLfloat fSize,   /* width/height of one tile     */
		      GLint iColumns,  /* number of columns to draw    */
		      GLint iRows      /* number of rows to draw       */)
{
	GLfloat fX = fStartX;
	GLfloat fY = fStartY;
	gboolean bColorToggle = TRUE;
	gboolean bStartedWith = TRUE;
	gint x;
	gint y;

	glBegin (GL_QUADS);
	for (x = 0; x < iColumns; x++)
	{
		bColorToggle = bStartedWith ? FALSE : TRUE;
		bStartedWith = bColorToggle;

		for (y = 0; y < iRows; y++)
		{
			if (bColorToggle)
				glColor4f (0.4f, 0.4f, 0.4f, 1.0f);
			else
				glColor4f (0.5f, 0.5f, 0.5f, 1.0f);

			bColorToggle = bColorToggle ? FALSE : TRUE;

			glVertex3f (fX,         fY,         0.0f);
			glVertex3f (fX,         fY + fSize, 0.0f);
			glVertex3f (fX + fSize, fY + fSize, 0.0f);
			glVertex3f (fX + fSize, fY,         0.0f);

			fY += fSize;
		}
		fX += fSize;
		fY = fStartY;
	}
	glEnd ();
}

/* this paints a nice white-greyish gradient; meant to be used in parallel
** projection */
void
opengl_paint_bg (GLfloat fStartX, /* x-coord of lower-left corner */
		 GLfloat fStartY, /* y-coord of lower-left corner */
		 GLfloat fWidth,  /* width of background          */
		 GLfloat fHeight  /* height of background         */)
{
	glBegin (GL_QUADS);

	/* top third, darker grey to white */
	glColor3f (0.85f, 0.85f, 0.85f);
	glVertex3f (fStartX, fStartY, 0.0f);
	glColor3f (0.85f, 0.85f, 0.85f);
	glVertex3f (fStartX + fWidth, fStartY, 0.0f);
	glColor3f (1.0f, 1.0f, 1.0f);
	glVertex3f (fStartX + fWidth, fStartY + fHeight / 3.0f, 0.0f);
	glColor3f (1.0f, 1.0f, 1.0f);
	glVertex3f (fStartX, fStartY + fHeight / 3.0f, 0.0f);

	/* middle third, just plain white */
	glColor3f (1.0f, 1.0f, 1.0f);
	glVertex3f (fStartX, fStartY + fHeight / 3.0f, 0.0f);
	glVertex3f (fStartX + fWidth, fStartY + fHeight / 3.0f, 0.0f);
	glVertex3f (fStartX + fWidth, fStartY + 2.0f * fHeight / 3.0f, 0.0f);
	glVertex3f (fStartX, fStartY + 2.0f * fHeight / 3.0f, 0.0f);

	/* bottom third, white to lighter grey */
	glColor3f (1.0f, 1.0f, 1.0f);
	glVertex3f (fStartX, fStartY + 2.0f * fHeight / 3.0f, 0.0f);
	glColor3f (1.0f, 1.0f, 1.0f);
	glVertex3f (fStartX + fWidth, fStartY + 2.0f * fHeight / 3.0f, 0.0f);
	glColor3f (0.62f, 0.66f, 0.69f);
	glVertex3f (fStartX + fWidth, fStartY + fHeight, 0.0f);
	glColor3f (0.62f, 0.66f, 0.69f);
	glVertex3f (fStartX, fStartY + fHeight, 0.0f);
	glEnd ();
}

/* the initial setup for GL, setting some required states, position the camera,
** setup shaders, switch on vsync, load/create textures */
void
opengl_init (GLint iWidth,
	     GLint iHeight)
{
	GLfloat afLightDiffuse[4] = {0.76f, 0.75f, 0.65f, 1.0f};

	/* set some GL-states */
	glClearColor (0.0f, 0.0f, 1.0f, 1.0f);
	glColor4f (1.0f, 1.0f, 1.0f, 0.0f);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, 1);

	/* "switch on the light" */
	glEnable (GL_LIGHTING);
	glEnable (GL_LIGHT0);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, afLightDiffuse);

	/* setup "camera" */
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	gluLookAt (0.0f, 0.0f, Z_NEAR,
		   0.0f, 0.0f, 0.0f,
		   0.0f, 1.0f, 0.0f);
	glTranslatef (0.0f, 0.0f, -Z_NEAR);

	/* try to find and initialize programmable shading capabilities */
	if (setup_shader_callbacks ())
	{
		g_bHasShaders = TRUE;
		glEnable (GL_FRAGMENT_PROGRAM_ARB);
		g_auiFragProgs[0] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "solid.frag");
		g_auiFragProgs[1] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "texture-2d.frag");
		g_auiFragProgs[2] = upload_program (GL_FRAGMENT_PROGRAM_ARB,
						    "texture-rect.frag");
	}

	/* switch on sync-to-vblank*/
	opengl_enable_vsync (TRUE);

	/* load and setup needed texture-resources */
	g_auiTextureIds[0] = opengl_upload_texture (GL_TEXTURE_2D,
						    "zebra.png",
						    TRUE);
	g_auiTextureIds[1] = opengl_upload_texture (GL_TEXTURE_2D,
						    "zebra.png",
						    FALSE);
	g_auiTextureIds[2] = opengl_upload_texture (GL_TEXTURE_2D,
						    "zebra-trimmed.png",
						    FALSE);

	g_apTextTextures[0] =  opengl_upload_text_texture ("trimmed in Gimp",
							   iWidth,
							   iHeight);
	g_apTextTextures[1] =  opengl_upload_text_texture ("trimmed with cairo",
							   iWidth,
							   iHeight);
	g_apTextTextures[2] =  opengl_upload_text_texture ("untrimmed",
							   iWidth,
							   iHeight);
}

/* needed to ensure proper resize-behaviour of the GL-context if the window is
** altered in size by the user */
void
opengl_resize (GLsizei iWidth, /* new width to adjust the GL-context to  */
	       GLsizei iHeight /* new height to adjust the GL-context to */)
{
	glViewport (0, 0, iWidth, iHeight);
	opengl_change_projection (TRUE, iWidth, iHeight);
	glClear (GL_COLOR_BUFFER_BIT);

	/* for the extreme rare case someone might try to compile this under
	** windows (you'll have to do some work on opengl_enable_vsync() then)
	** I've added the following calls; under Linux/Unix they are generally
	** not needed, but under Windows a resize of the GL-context invalidates
	** any texture-object, thus it has to be recreated (a little sidenote
	** regarding this issue: remember id-software games, which are OpenGL-
	** -based... they regenerate all their game resources if you alter the 
	** game-resolution) */
	glDeleteTextures (3, g_auiTextureIds);
	g_auiTextureIds[0] = opengl_upload_texture (GL_TEXTURE_2D,
						    "zebra.png",
						    TRUE);
	g_auiTextureIds[1] = opengl_upload_texture (GL_TEXTURE_2D,
						    "zebra.png",
						    FALSE);
	g_auiTextureIds[2] = opengl_upload_texture (GL_TEXTURE_2D,
						    "zebra_trimmed.png",
						    FALSE);

	g_slice_free1 (sizeof (Texture), (gpointer) g_apTextTextures[0]);
	g_slice_free1 (sizeof (Texture), (gpointer) g_apTextTextures[1]);
	g_slice_free1 (sizeof (Texture), (gpointer) g_apTextTextures[2]);
	g_apTextTextures[0] =  opengl_upload_text_texture ("trimmed in Gimp",
							  iWidth,
							  iHeight);
	g_apTextTextures[1] =  opengl_upload_text_texture ("trimmed with cairo",
							  iWidth,
							  iHeight);
	g_apTextTextures[2] =  opengl_upload_text_texture ("untrimmed",
							  iWidth,
							  iHeight);
}

/* the main action takes place here once everything is setup; this is what gets
** called very time we force a refresh/redraw of the window */
void
opengl_draw (GLsizei iWidth, /* current width of the GL-context  */
	     GLsizei iHeight /* current height of the GL-context */)
{
	static GLfloat fAngle = 0.0f;
	GLfloat afMaterialDiffuseFrontFull[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	GLfloat afMaterialDiffuseBackFull[4] = {0.0f, 0.16f, 0.31f, 1.0f};
	GLfloat afMaterialDiffuseFrontOff[4] = {1.0f, 1.0f, 1.0f, 0.0f};
	GLfloat afMaterialDiffuseBackOff[4] = {0.0f, 0.16f, 0.31f, 0.0f};
	GLfloat afMaterialDiffuseFrontQuarter[4] = {1.0f, 1.0f, 1.0f, 0.25f};
	GLfloat afMaterialDiffuseBackQuarter[4] = {0.0f, 0.16f, 0.31f, 0.25f};

	glClear (GL_COLOR_BUFFER_BIT);

	/* disable texturing and lighting and draw the background using the
	** first shader */
	opengl_change_projection (TRUE, iWidth, iHeight);
	glDisable (GL_TEXTURE_RECTANGLE_ARB);
	glDisable (GL_TEXTURE_2D);
	glDisable (GL_LIGHTING);
	glBindProgramARB (GL_FRAGMENT_PROGRAM_ARB, g_auiFragProgs[0]);
	opengl_paint_bg (-iWidth/2.0f, -iHeight/2.0f, iWidth, iHeight);

	/* enable texturing and lighting and draw the spinning image with the
	** other shader */
	opengl_change_projection (FALSE, iWidth, iHeight);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_LIGHTING);
	glBindProgramARB (GL_FRAGMENT_PROGRAM_ARB, g_auiFragProgs[1]);
	fAngle += 1.0f;

	glBindTexture (GL_TEXTURE_2D, g_auiTextureIds[2]);
	glPushMatrix ();
	glTranslatef (-3.5f, 0.0f, -5.0f);
	glRotatef (fAngle, 0.0f, 1.0f, 0.0f);
	glBegin (GL_QUADS);
	/* left example  */
	glNormal3f (0.0f, 0.0f, 1.0f);
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontFull);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackFull);
	glTexCoord2f (0.0f, 1.0f);
	glVertex3f (-1.5f, -1.5375f, 0.0f);
	glTexCoord2f (1.0f, 1.0f);
	glVertex3f ( 1.5f, -1.5375f, 0.0f);
	glTexCoord2f (1.0f, 0.0f);
	glVertex3f ( 1.5f,  1.4625f, 0.0f);
	glTexCoord2f (0.0f, 0.0f);
	glVertex3f (-1.5f,  1.4625f, 0.0f);
	/* left example's reflection */
	glColor4f (1.0f, 1.0f, 1.0f, 0.0f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontOff);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackOff);
	glTexCoord2f (0.0f, 0.5f);
	glVertex3f (-1.5f, -2.9625f, 0.0f);
	glTexCoord2f (1.0f, 0.5f);
	glVertex3f ( 1.5f, -2.9625f, 0.0f);
	glColor4f (1.0f, 1.0f, 1.0f, 0.25f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontQuarter);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackQuarter);
	glTexCoord2f (1.0f, 1.0f);
	glVertex3f ( 1.5f, -1.4625f, 0.0f);
	glTexCoord2f (0.0f, 1.0f);
	glVertex3f (-1.5f, -1.4625f, 0.0f);
	glEnd ();
	glPopMatrix ();

	glBindTexture (GL_TEXTURE_2D, g_auiTextureIds[1]);
	glPushMatrix ();
	glTranslatef (3.5f, 0.0f, -5.0f);
	glRotatef (fAngle, 0.0f, 1.0f, 0.0f);
	glBegin (GL_QUADS);
	/* right example */
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontFull);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackFull);
	glTexCoord2f (0.0f, 1.0f);
	glVertex3f (-1.5f, -1.5f, 0.0f);
	glTexCoord2f (1.0f, 1.0f);
	glVertex3f ( 1.5f, -1.5f, 0.0f);
	glTexCoord2f (1.0f, 0.0f);
	glVertex3f ( 1.5f,  1.5f, 0.0f);
	glTexCoord2f (0.0f, 0.0f);
	glVertex3f (-1.5f,  1.5f, 0.0f);
	/* right example's reflection */
	glColor4f (1.0f, 1.0f, 1.0f, 0.0f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontOff);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackOff);
	glTexCoord2f (0.0f, 0.5f);
	glVertex3f (-1.5f, -3.0f, 0.0f);
	glTexCoord2f (1.0f, 0.5f);
	glVertex3f ( 1.5f, -3.0f, 0.0f);
	glColor4f (1.0f, 1.0f, 1.0f, 0.25f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontQuarter);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackQuarter);
	glTexCoord2f (1.0f, 1.0f);
	glVertex3f ( 1.5f, -1.5f, 0.0f);
	glTexCoord2f (0.0f, 1.0f);
	glVertex3f (-1.5f, -1.5f, 0.0f);
	glEnd ();
	glPopMatrix ();

	glBindTexture (GL_TEXTURE_2D, g_auiTextureIds[0]);
	glPushMatrix ();
	glTranslatef (0.0f, 0.0f, -5.0f);
	glRotatef (fAngle, 0.0f, 1.0f, 0.0f);
	glBegin (GL_QUADS);
	/* middle example */
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontFull);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackFull);
	glTexCoord2f (0.0f, 1.0f);
	glVertex3f (-1.5f, -1.525f, 0.0f);
	glTexCoord2f (1.0f, 1.0f);
	glVertex3f ( 1.5f, -1.525f, 0.0f);
	glTexCoord2f (1.0f, 0.0f);
	glVertex3f ( 1.5f,  1.475f, 0.0f);
	glTexCoord2f (0.0f, 0.0f);
	glVertex3f (-1.5f,  1.475f, 0.0f);
	/* middle example's reflection */
	glColor4f (1.0f, 1.0f, 1.0f, 0.0f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontOff);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackOff);
	glTexCoord2f (0.0f, 0.5f);
	glVertex3f (-1.5f, -2.75f, 0.0f);
	glTexCoord2f (1.0f, 0.5f);
	glVertex3f ( 1.5f, -2.75f, 0.0f);
	glColor4f (1.0f, 1.0f, 1.0f, 0.25f);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, afMaterialDiffuseFrontQuarter);
	glMaterialfv (GL_BACK, GL_DIFFUSE, afMaterialDiffuseBackQuarter);
	glTexCoord2f (1.0f, 1.0f);
	glVertex3f ( 1.5f, -1.475f, 0.0f);
	glTexCoord2f (0.0f, 1.0f);
	glVertex3f (-1.5f, -1.475f, 0.0f);
	glEnd ();
	glPopMatrix ();

	/* draw the text-labels */
	opengl_change_projection (TRUE, iWidth, iHeight);
	glDisable (GL_TEXTURE_2D);
	glDisable (GL_LIGHTING);
	glBindProgramARB (GL_FRAGMENT_PROGRAM_ARB, g_auiFragProgs[2]);

	glEnable (g_apTextTextures[0]->iTarget);
	glBindTexture (g_apTextTextures[0]->iTarget,
		       g_apTextTextures[0]->uiTextureId);
	glPushMatrix ();
	glTranslatef (- iWidth / 3 - g_apTextTextures[0]->fWidth / 2.0f, -1.0f * (GLfloat) iHeight / 2 + 20.0f, 0.0f);
	glBegin (GL_QUADS);
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f, g_apTextTextures[0]->fHeight);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glTexCoord2f (g_apTextTextures[0]->fWidth,
		      g_apTextTextures[0]->fHeight);
	glVertex3f (g_apTextTextures[0]->fWidth, 0.0f, 0.0f);
	glTexCoord2f (g_apTextTextures[0]->fWidth, 0.0f);
	glVertex3f (g_apTextTextures[0]->fWidth,
		    g_apTextTextures[0]->fHeight,
		    0.0f);
	glTexCoord2f (0.0f, 0.0f);
	glVertex3f (0.0f, g_apTextTextures[0]->fHeight, 0.0f);
	glEnd ();
	glPopMatrix ();
	glDisable (g_apTextTextures[0]->iTarget);

	glEnable (g_apTextTextures[1]->iTarget);
	glBindTexture (g_apTextTextures[1]->iTarget,
		       g_apTextTextures[1]->uiTextureId);
	glPushMatrix ();
	glTranslatef (g_apTextTextures[1]->fWidth / -2.0f, -1.0f * (GLfloat) iHeight / 2 + 20.0f, 0.0f);
	glBegin (GL_QUADS);
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f, g_apTextTextures[1]->fHeight);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glTexCoord2f (g_apTextTextures[1]->fWidth,
		      g_apTextTextures[1]->fHeight);
	glVertex3f (g_apTextTextures[1]->fWidth, 0.0f, 0.0f);
	glTexCoord2f (g_apTextTextures[1]->fWidth, 0.0f);
	glVertex3f (g_apTextTextures[1]->fWidth,
		    g_apTextTextures[1]->fHeight,
		    0.0f);
	glTexCoord2f (0.0f, 0.0f);
	glVertex3f (0.0f, g_apTextTextures[1]->fHeight, 0.0f);
	glEnd ();
	glPopMatrix ();
	glDisable (g_apTextTextures[1]->iTarget);

	glEnable (g_apTextTextures[2]->iTarget);
	glBindTexture (g_apTextTextures[2]->iTarget,
		       g_apTextTextures[2]->uiTextureId);
	glPushMatrix ();
	glTranslatef (iWidth / 3 - g_apTextTextures[2]->fWidth / 2.0f, -1.0f * (GLfloat) iHeight / 2 + 20.0f, 0.0f);
	glBegin (GL_QUADS);
	glColor4f (1.0f, 1.0f, 1.0f, 1.0f);
	glTexCoord2f (0.0f, g_apTextTextures[2]->fHeight);
	glVertex3f (0.0f, 0.0f, 0.0f);
	glTexCoord2f (g_apTextTextures[2]->fWidth,
		      g_apTextTextures[2]->fHeight);
	glVertex3f (g_apTextTextures[2]->fWidth, 0.0f, 0.0f);
	glTexCoord2f (g_apTextTextures[2]->fWidth, 0.0f);
	glVertex3f (g_apTextTextures[2]->fWidth,
		    g_apTextTextures[2]->fHeight,
		    0.0f);
	glTexCoord2f (0.0f, 0.0f);
	glVertex3f (0.0f, g_apTextTextures[2]->fHeight, 0.0f);
	glEnd ();
	glPopMatrix ();
	glDisable (g_apTextTextures[2]->iTarget);
}

/* just a tiny bit of resource cleaning needed to happen at the end */
void
opengl_cleanup (void)
{
	if (g_bHasShaders)
		glDeleteProgramsARB (3, g_auiFragProgs);

	glDeleteTextures (3, g_auiTextureIds);

	g_slice_free1 (sizeof (Texture), (gpointer) g_apTextTextures[0]);
	g_slice_free1 (sizeof (Texture), (gpointer) g_apTextTextures[1]);
	g_slice_free1 (sizeof (Texture), (gpointer) g_apTextTextures[2]);
}
